#! /usr/bin/env python

import tomoesnml
import pandas as pd
import numpy as np
import psycopg2
import warnings
import astropy
import shutil
import random
import os,sys
import math
import time
from timeout_decorator import timeout, TimeoutError



def main(root_directory, source_id_list):
    ##---------- Initial setting

    directory_new_path  = '/lustre/tomoesn/raw'
    directory_temp_path = '/lustre/tomoesn/reference'
    psql_login = 'host=gwdata dbname=supernova user=tomoesn password='

    directory_out_path = root_directory

    warnings.simplefilter('ignore', astropy.wcs.wcs.FITSFixedWarning)


    
    with psycopg2.connect(psql_login) as connection:
        ra_list, dec_list, raw_id_list, path_list = inquire_source_sub(connection, source_id_list, directory_new_path)


    print 'sourceId, ra, dec, rawId, path'
    for source_id, source_ra, source_dec, raw_id, new_fits_path in zip(source_id_list, ra_list, dec_list, raw_id_list, path_list):
        print source_id, source_ra, source_dec, raw_id, new_fits_path
    print ''
    
    number_of_processed_files = 0
    number_of_successed_files = 0
    num_of_ArtificialReals    = 0
    num_of_Artifacts          = 0

    for source_id, source_ra, source_dec, raw_id, new_fits_path in zip(source_id_list, ra_list, dec_list, raw_id_list, path_list):
        with psycopg2.connect(psql_login) as connection:
            with connection.cursor() as cur: cur.execute('SET statement_timeout TO 60000') # 60 [sec]

            number_of_processed_files += 1

            start_time = time.time()
            part_elapsed_time_dict = {}

            if (new_fits_path is None) or (os.path.isfile(new_fits_path) == 0): # Skip NULL and unknown file path
                print ' [WARN] Not exist file. -> {}'.format(new_fits_path)
                connection.commit()
                continue

            keywords = os.path.splitext(os.path.basename(new_fits_path))[0]

            print '  File name           : ' + keywords
            print '  raw id              : ' + str(raw_id)

            fwhm, zero_mag, lim_mag, num_of_sources = inquire_sources(connection, raw_id) # Check rawId
            print '  fwhm, zero mag      : ' + str(fwhm) + ', ' + str(zero_mag)        

            # ===============================================================
            # Detect Source from new image
            # ===============================================================

            part_start_time = time.time()

            new_sex = tomoesnml.SourceExtractor(new_fits_path,
                                                detect_min_area  = 5,
                                                detect_threshold = 1.8,
                                                mag_zero_point   = 22.5,
                                                pixel_scale      = 1.19,
                                                seeing_fwhm      = 3.0)

            if new_sex.dataframe is None:
                print ' [WARN] ' + keywords + ' : ' + 'No columns to parse from file !'
                print ''
                tomoesnml.remove_files(keywords+'*')
                connection.commit()
                continue

            try:
                table_name = 'hamasaki_table_' + str(os.getpid())
                panstarrs_dataframe = get_panstarrs_data(connection, new_sex.dataframe,
                                                         temp_table_name=table_name,
                                                         max_num_of_inquire_objects=4000)
            except psycopg2.extensions.QueryCanceledError: # TimeoutError:
                print ' [WARN] ' + keywords + ' : ' + 'Timeout error of inquire stars !'
                print ''
                tomoesnml.remove_files(keywords+'*')
                connection.commit()
                continue
            except:
                print ' [WARN] ' + keywords + ' : ' + 'Skip because occured exception !'
                print ''
                tomoesnml.remove_files(keywords+'*')
                connection.commit()
                continue

            # panstarrs_dataframe.columns: ['TYPE_STAR','OBJ_INFO_FLAG','PS_STAR','DIST_PS1']
            new_sex.dataframe = new_sex.dataframe.join(panstarrs_dataframe) 
            # new_sex.dataframe.columns: ['NUMBER', 'X_IMAGE', 'Y_IMAGE', 'ALPHA_J2000',
            #                             'DELTA_J2000', 'MAG_AUTO', 'MAGERR_AUTO', 'A_IMAGE',
            #                             'B_IMAGE', 'THETA_IMAGE', 'ELONGATION', 'FWHM_IMAGE',
            #                             'FLUX_AUTO', 'FLUXERR_AUTO', 'FLUX_MAX', 'CLASS_STAR',
            #                             'TYPE_STAR', 'OBJ_INFO_FLAG', 'PS_STAR', 'DIST_PS1']

            tomoesnml.MakeReg(keywords+'_new.reg',
                              new_sex.dataframe['X_IMAGE'].values,
                              new_sex.dataframe['Y_IMAGE'].values,
                              values=new_sex.dataframe['MAG_AUTO'].values,
                              numbers=new_sex.dataframe['PS_STAR'].values,
                              color='green')

            part_elapsed_time_dict['Detect Sources from new-image'] = time.time() - part_start_time

            # ----------------------------------------------------
            # new fits path        : new_fits_path
            # embed new fits path  : embed_new_fits_path
            # embed sub fits path  : embed_sub_fits_path
            # temp fits path       : temp_fits_path
            # temp remap fits path : temp_remap_fits_path
            # ----------------------------------------------------

            # ===============================================================
            # Make Artificial
            # ===============================================================

            part_start_time = time.time()
            psf_map_size = 14 # Length of ArtificialReal PSF map : psf_map_size*2 + 1

            
            part_elapsed_time_dict['Make Artificial'] = time.time() - part_start_time
            
            # ===============================================================
            # Subtract fits image
            # ===============================================================

            part_start_time = time.time()

            temp_remap_fits_path = keywords+'_temp_remap.fits'
            embed_sub_fits_path  = keywords+'_sub.fits'
            
            flag, temp_fits_path = tomoesnml.imsub(connection, new_fits_path,
                                                   temp_remap_fits_name=temp_remap_fits_path,
                                                   subtracted_fits_name=embed_sub_fits_path)
            if flag==1:
                print ' [WARN] ' + keywords + ' : ' + 'Template image does not exist !'
                print ''
                tomoesnml.remove_files(keywords+'*')
                connection.commit()
                continue

            part_elapsed_time_dict['Subtract fits image'] = time.time() - part_start_time
            shutil.copy(new_fits_path, keywords+'_new.fits')
            shutil.copy(temp_fits_path, keywords+'_temp.fits')

            part_elapsed_time_dict['Subtract fits image'] = time.time() - part_start_time

            # ===============================================================
            # Detect Source from sub image
            # ===============================================================

            part_start_time = time.time()

            sub_sex = tomoesnml.SourceExtractor(embed_sub_fits_path,
                                                detect_min_area  = 2,
                                                detect_threshold = 3.,
                                                mag_zero_point   = 20.0,
                                                pixel_scale      = 1.19,
                                                seeing_fwhm      = 4.0)

            if sub_sex.dataframe is None:
                print ' [WARN] ' + keywords + ' : ' + 'No columns to parse from file !'
                print ''
                tomoesnml.remove_files(keywords+'*')
                connection.commit()
                continue

            part_elapsed_time_dict['Detect Sources from sub-image'] = time.time() - part_start_time

            # ===============================================================
            # Filter source & Add value
            # ===============================================================

            part_start_time = time.time()
            

            panstarrs_catalog_df = new_sex.dataframe[(new_sex.dataframe['PS_STAR'] != 0)][['ALPHA_J2000','DELTA_J2000','OBJ_INFO_FLAG','PS_STAR']]
            tomoesnml.MakeReg(keywords+'_pscatalog.reg',
                              panstarrs_catalog_df['ALPHA_J2000'].values,
                              panstarrs_catalog_df['DELTA_J2000'].values,
                              values=panstarrs_catalog_df['OBJ_INFO_FLAG'].values,
                              numbers=panstarrs_catalog_df['PS_STAR'].values,
                              color='pink',coordsys='FK5',circle_radius=0.005,y_desc=0.003)


            sub_sex.dataframe[['DIST_PS1','OBJ_INFO_FLAG','PS_STAR']] = sub_sex.dataframe.apply(check_panstarrs_catalog,
                                                                                                panstarrs_catalog_df=panstarrs_catalog_df,
                                                                                                axis=1)
            sub_sex.dataframe['S2N'] = sub_sex.dataframe.apply(calculate_signal_to_noise_ratio, axis=1)
            sub_sex.dataframe['FITS_FILE_NAME']    = keywords
            sub_sex.dataframe['SQL_INFO_RAW_ID']   = raw_id
            sub_sex.dataframe['SQL_INFO_FWHM']     = fwhm
            sub_sex.dataframe['SQL_INFO_ZERO_MAG'] = zero_mag
            sub_sex.dataframe['SQL_INFO_LIM_MAG']  = lim_mag

            check_too_close_sources(sub_sex.dataframe, flag_name='TOO_CLOSE')

            with tomoesnml.FilterSource(temp_fits_path, temp_remap_fits_path, embed_sub_fits_path) as filter_process:
                sub_sex.dataframe['VALUE_ZERO_REF'] = sub_sex.dataframe.apply(filter_process.check_non_ref,    axis=1)
                sub_sex.dataframe['SATURATION']     = sub_sex.dataframe.apply(filter_process.check_saturation, axis=1)
                sub_sex.dataframe['MASK']           = sub_sex.dataframe.apply(filter_process.check_mask,       axis=1)
                sub_sex.dataframe['EDGE']           = sub_sex.dataframe.apply(filter_process.check_edge,       axis=1)

            tomoesnml.MakeReg(keywords+'_sub.reg',
                              sub_sex.dataframe['X_IMAGE'].values,
                              sub_sex.dataframe['Y_IMAGE'].values,
                              values=sub_sex.dataframe['MAG_AUTO'].values,
                              numbers=sub_sex.dataframe['PS_STAR'].values,
                              color='cyan')

            part_elapsed_time_dict['Filter source & Add value'] = time.time() - part_start_time

            # ===============================================================
            # Output
            # ===============================================================

            # ----------------------------------------------------
            # new fits path        : new_fits_path
            # embed new fits path  : embed_new_fits_path
            # embed sub fits path  : embed_sub_fits_path
            # temp fits path       : temp_fits_path
            # temp remap fits path : temp_remap_fits_path
            # ----------------------------------------------------

            part_start_time = time.time()


            part_elapsed_time_dict['Output'] = time.time() - part_start_time

            tomoesnml.MakeReg(keywords+'_target_source_sub.reg',
                              np.array([source_ra]),
                              np.array([source_dec]),
                              color='purple',coordsys='FK5',
                              circle_radius=3.0/3600,y_desc=6.0/3600)

            # ===============================================================
            # Move files
            # ===============================================================

            # Check rawId
            with open(keywords+'_info.dat', 'w') as f:
                f.write("# fwhm, zero mag, lim mag, num of sources\n")
                f.write("{}, {}, {}, {}\n".format(fwhm, zero_mag, lim_mag, num_of_sources))

            # Move files and Remove file
            directory_out_path2 = os.path.join(directory_out_path, str(source_id))
            if not os.path.isdir(directory_out_path2): os.makedirs(directory_out_path2)
            tomoesnml.move_files(keywords+'*', directory_out_path2)

            # ===============================================================
            # Finish
            # ===============================================================

            # Output processing time
            elapsed_time = time.time() - start_time
            print '  Processed  number         : {} (included failure {} )'.format(number_of_successed_files+1, number_of_processed_files)
            print '  Processing time           : {}'.format(elapsed_time)
            for part_time_key in part_elapsed_time_dict: # for print measurement time
                #print ' {:<40} : {}'.format(part_time_key, part_elapsed_time_dict[part_time_key])
                pass
            print ''
            number_of_successed_files += 1

            connection.commit()

    print '  Number of processed files  : {} (included failure {} )'.format(number_of_successed_files, number_of_processed_files)



def inquire_source_sub(connection, source_id_list, directory_new_path):
    source_sub_ra_list = []
    source_sub_dec_list = []
    source_sub_raw_id_list = []
    source_sub_path_list = []
    with connection.cursor() as cur:
        for source_id in source_id_list:
            query = ('SELECT source_sub."ra", source_sub."dec", raw."rawId", raw."fitsName" '+
                     'FROM source_sub INNER JOIN raw ON source_sub."rawId" = raw."rawId" WHERE source_sub."sourceId" = {}'.format(source_id))
            cur.execute(query)
            responses = cur.fetchone()
            source_sub_ra_list.append(responses[0])
            source_sub_dec_list.append(responses[1])
            source_sub_raw_id_list.append(responses[2])
            source_sub_path_list.append(os.path.join(directory_new_path, responses[3]))
    return source_sub_ra_list, source_sub_dec_list, source_sub_raw_id_list, source_sub_path_list



def inquire_new_fits(connection, target_det_id, date_min, date_max, directory_new_path='/lustre/tomoesn/raw', ml_project_name='Supernova Survey'):

    query = 'SELECT "rawId","expId","detId","fieldId","filter","fitsName","project","mjd" FROM raw WHERE "project" = \'{}\' AND "dateObs">=\'{}\' AND "dateObs"<=\'{}\' '.format(ml_project_name,date_min,date_max)
    with connection.cursor() as cur:
        try:
            cur.execute(query)
        except:
            time.sleep(10)
            cur.execute(query)
        row = cur.fetchall()

    if(row != None):
        rawId_list         = [data[0] for data in row]
        expId_list         = [data[1] for data in row]
        detId_list         = [data[2] for data in row]
        fieldId_list       = [data[3] for data in row]
        #filter_list       = [data[4] for data in row]
        fits_name_list     = [data[5] for data in row]
        #project_name_list = [data[6] for data in row]
        mjd_list           = [data[7] for data in row]

        fits_path_list = [os.path.join(directory_new_path, fits_name) for fits_name in fits_name_list]
        det_num_fits_data_list = [ [ fits_path, raw_id ]
                                   for fits_path, det_id, raw_id in zip(fits_path_list, detId_list, rawId_list)
                                   if det_id == target_det_id]
        return det_num_fits_data_list
    else:
        return None



def inquire_sources(connection, raw_id):
    query = 'SELECT "fwhm","zeromag","limmag","numofsources","sn5mag","sn5magsub" FROM analysis_stack WHERE "rawId" = \'{}\' '.format(raw_id)
    with connection.cursor() as cur:
        try:
            cur.execute(query)
        except:
            time.sleep(10)
            cur.execute(query)
        row = cur.fetchone()
    if(row != None):
        fwhm           = row[0]
        zero_mag       = row[1]
        lim_mag        = row[2]
        num_of_sources = row[3]
        sn5_mag        = row[4]
        sn5_mag_sub    = row[5]

        return fwhm, zero_mag, lim_mag, num_of_sources
    else:
        return None, None, None, None



def get_panstarrs_data(connection, dataframe, temp_table_name='hamasaki_table', max_num_of_inquire_objects=8000):

    def inquire_psql_database(ra_np, dec_np, temp_table_name):
        ra_dec_np       = np.stack([ra_np,dec_np], axis=-1)
        ra_dec_tuple    = tuple(map(tuple,ra_dec_np)) # ra_dec_tuple = ((ra1, dec1),(ra2, dec2)...(ra, dec))
        ra_dec_str      = ','.join(map(str, ra_dec_tuple))
        with connection.cursor() as cur:
            query = 'CREATE TEMPORARY TABLE {} (ra DOUBLE PRECISION, dec DOUBLE PRECISION)'.format(temp_table_name)
            cur.execute(query)

            # Insert (ra,dec)data to my_table
            query = 'INSERT INTO {} (ra,dec) VALUES {}'.format(temp_table_name, ra_dec_str)
            cur.execute(query)
        with connection.cursor('match_ps_table') as cur: # ---!!!CAUTION!!!--- Heavy Processing
            cur.itersize = 3000
            # Search ps1 catalog
            query = 'SELECT ss.* FROM {} AS t LEFT JOIN LATERAL (SELECT s."psId", s."objInfoFlag", q3c_dist(t.ra, t.dec, s.ra, s.dec) FROM ps_bright AS s WHERE q3c_join(t.ra, t.dec, s.ra, s.dec, 1./60) ORDER BY q3c_dist(t.ra, t.dec, s.ra, s.dec) ASC LIMIT 1 ) AS ss ON true'.format(temp_table_name)
            cur.execute(query)
            responses = cur.fetchall()
        return responses

    def get_ps_data(ps_info):
        match_rad   = 3./3600 # [arcsec]
        objInfoFlag = 0
        psstar      = 0
        dist_ps1    = 0
        type_star   = False
        if ps_info:
            psId,objInfoFlag,dist_ps1 = ps_info
            psstar = 1
            if dist_ps1 <= match_rad:
                if objInfoFlag:
                    for iobj,objInfo in enumerate(['FEW','POOR','ICRF_QSO','HERN_QSO_P60','HERN_QSO_P05',
                                                   'HERN_RRL_P60','HERN_RRL_P05','HERN_VARIABLE','TRANSIENT',
                                                   'HAS_SOLSYS_DET','MOST_SOLSYS_DET','LARGE_PM','RAW_AVE',
                                                   'FIT_AVE','FIT_PM','FIT_PAR','USE_AVE','USE_PM','USE_PAR',
                                                   'NO_MEAN_ASTROM','STACK_FOR_MEAN','MEAN_FOR_STACK','BAD_PM',
                                                   'EXT','EXT_ALT','GOOD','GOOD_ALT','GOOD_STACK','BEST_STACK',
                                                   'SUSPECT_STACK','BAD_STACK']):
                        if objInfoFlag & 2**iobj:
                            if objInfo == 'EXT':
                                psstar = 2
                else:
                    psstar = 0
                    dist_ps1 = 0
                    objInfoFlag = 0
            else:
                psstar = 0
                dist_ps1 = 0
                objInfoFlag = 0
        if psstar == 1 :
            type_star = True
        return [type_star, objInfoFlag, psstar, dist_ps1]

    ra_np  = dataframe['ALPHA_J2000'].values
    dec_np = dataframe['DELTA_J2000'].values
    if ra_np.shape[0] > max_num_of_inquire_objects: raise Exception('Too many objects : {}'.format(ra_np.shape[0]))
    try: responses = inquire_psql_database(ra_np, dec_np, temp_table_name)
    except: raise
    else:
        ps_list = [get_ps_data(res) for res in responses]
        ps_df = pd.DataFrame(ps_list, columns=['TYPE_STAR','OBJ_INFO_FLAG','PS_STAR','DIST_PS1'])
        return ps_df



def check_artificial_real(pd_series, artificial_real_df):
    flag = False
    pixel_gap = 3.
    for x0_image,y0_image in zip(artificial_real_df['X_IMAGE'].values, artificial_real_df['Y_IMAGE'].values):
        if( (pixel_gap > abs(pd_series['X_IMAGE'] - x0_image)) and
            (pixel_gap > abs(pd_series['Y_IMAGE'] - y0_image)) ):
            flag = True
    return flag



def check_panstarrs_catalog(pd_series, panstarrs_catalog_df):
    # Use panstarrs_catalog_df            
    match_rad   = 3.0/3600 #[arcsec]/3600
    ra          = pd_series['ALPHA_J2000']
    dec         = pd_series['DELTA_J2000']

    # Search from panstarrs catalog
    ps_match_list = []
    for ps_catalog_tuple in panstarrs_catalog_df.itertuples(name=None):
        ps_ra  = ps_catalog_tuple[1]
        ps_dec = ps_catalog_tuple[2]
        dist_ps1_square = (ra - ps_ra)**2 + (dec - ps_dec)**2
        if dist_ps1_square < match_rad**2:
            ps_match_list.append([ np.sqrt(dist_ps1_square), ps_catalog_tuple[3], ps_catalog_tuple[4] ])
    if len(ps_match_list) != 0:
        ps_catalog_data = min(ps_match_list)
    else:
        ps_catalog_data = [None, None, 0]
    return pd.Series(ps_catalog_data)



def calculate_signal_to_noise_ratio(pd_series):
    s2n = None
    if isinstance(pd_series['FLUX_AUTO'], float) and isinstance(pd_series['FLUXERR_AUTO'], float):
        try:
            s2n = pd_series['FLUX_AUTO'] / pd_series['FLUXERR_AUTO']
        except:
            s2n = None
    return s2n



def check_too_close_sources(data_frame, flag_name='TOO_CLOSE'):
    radius = 14 + 3
    for i in range(data_frame.shape[0]):
        for j in range(i+1, data_frame.shape[0]):
            x = data_frame.at[data_frame.index[i],'X_IMAGE'] - data_frame.at[data_frame.index[j],'X_IMAGE']
            y = data_frame.at[data_frame.index[i],'Y_IMAGE'] - data_frame.at[data_frame.index[j],'Y_IMAGE']
            if (x*x + y*y) < (radius+1)*(radius+1) :
                data_frame.at[data_frame.index[i],flag_name] = True
                data_frame.at[data_frame.index[j],flag_name] = True



if __name__=='__main__':
    print 'Start.'
    root_directory = 'source'
    source_id_list = [919523713, 932293667, 1009976962]
    main(root_directory,source_id_list)
    print 'Finish.'
    sys.exit(0)
