#! /usr/bin/env python

import tomoesnml
import numpy as np
import pandas as pd

import astropy.io.fits
import astropy.coordinates
import astropy.time
import astropy.units

import psycopg2
import random
import os,sys
import shutil
import math
import time
import csv
from datetime import datetime
from pytz import timezone

import matplotlib.pyplot as plt


def main(date_min, date_max, target_num_of_sample_points=10000, project_name='Supernova Survey', output_path='./'):

    psql_login = 'host=gwdata dbname=supernova user=tomoesn password='
    keywords = date_min + '_' + date_max
    if not os.path.isdir(output_path): os.makedirs(output_path)

    np.random.seed(0)

    print 'Date: {} --- {}'.format(date_min, date_max)
    print 'Number of Data: {}'.format(target_num_of_sample_points)



    # ----------------------------------------------------------------------------------------
    # get source_sub data

    start_time = time.time()

    with psycopg2.connect(psql_login) as connection:
        with connection.cursor('inquire_source_sub_data') as cur:
            cur.itersize = 10000
            query = ('SELECT source_sub."sourceId", source_sub."ra", source_sub."dec" ' +
                     'FROM raw INNER JOIN source_sub ON raw."rawId" = source_sub."rawId" WHERE raw."project"=\'{}\' AND raw."dateObs" > \'{}\' AND raw."dateObs" < \'{}\' '.format(project_name, date_min, date_max) )
            cur.execute(query)
            responses = cur.fetchall()
        connection.commit()
    if(responses != None):
        source_sub_df = pd.DataFrame(responses, columns=['source_id', 'ra', 'dec'])
        output_file_path = os.path.join(output_path, 'source_sub_'+keywords+'.csv')
        source_sub_df.to_csv(output_file_path, header=False, index=False)
    else:
        source_sub_df = None

    elapsed_time = time.time() - start_time
    print 'P1 Time: {} hour'.format(elapsed_time/3600)

    # ----------------------------------------------------------------------------------------
    # choice coordinate point at random

    start_time = time.time()

    num_of_sample_points = target_num_of_sample_points

    ra = random_coordinate(0,120,num_of_sample_points*36/52)
    dec = random_coordinate(-5,70,num_of_sample_points*36/52)
    radius = 60.0/3600
    region1_df_list = [source_sub_df.query('(@radec[0]-@radius < ra < @radec[0]+@radius) and (@radec[1]-@radius < dec < @radec[1]+@radius)')
                       for radec in np.stack([ra,dec], axis=1)]

    ra = random_coordinate(120,170,num_of_sample_points*7/52)
    dec = random_coordinate(-5,30,num_of_sample_points*7/52)
    radius = 60.0/3600
    region2_df_list = [source_sub_df.query('(@radec[0]-@radius < ra < @radec[0]+@radius) and (@radec[1]-@radius < dec < @radec[1]+@radius)')
                       for radec in np.stack([ra,dec], axis=1)]

    ra = random_coordinate(330,360,num_of_sample_points*9/52)
    dec = random_coordinate(-5,70,num_of_sample_points*9/52)
    radius = 60.0/3600
    region3_df_list = [source_sub_df.query('(@radec[0]-@radius < ra < @radec[0]+@radius) and (@radec[1]-@radius < dec < @radec[1]+@radius)')
                       for radec in np.stack([ra,dec], axis=1)]

    region1_df_list.extend(region2_df_list)
    region1_df_list.extend(region3_df_list)
    sample_source_sub_df = pd.concat(region1_df_list,axis=0)

    id_ra_dec_csv_file_name = 'sample_source_sub_' + keywords + '.csv'
    id_ra_dec_csv_file_path = os.path.join(output_path, id_ra_dec_csv_file_name)

    sample_source_sub_df.to_csv(id_ra_dec_csv_file_path, header=False, index=False)

    elapsed_time = time.time() - start_time
    print 'P2 Time: {} hour'.format(elapsed_time/3600)

    sample_source_sub_df.reset_index(drop=True, inplace=True)

    # ----------------------------------------------------------------------------------------
    # filter nearby Pan-STARRS star

    start_time = time.time()

    with psycopg2.connect(psql_login) as connection:
        nearest_ps1_df = inquire_nearest_ps_star(connection, id_ra_dec_csv_file_path, temp_table_name='hamasaki_ps_star')
        # ps1_match_df.columns = ['psId','objInfoFlag','rmag','dist','psType']

    nearest_ps1_df.reset_index(drop=True, inplace=True)
    nearest_ps1_df['variable_star_flag'] = nearest_ps1_df.apply(check_variable_star, axis=1)

    sample_variable_df = sample_source_sub_df[nearest_ps1_df['variable_star_flag']==False]

    id_ra_dec_csv_file_name = 'sample_variable_' + keywords + '.csv'
    id_ra_dec_csv_file_path = os.path.join(output_path, id_ra_dec_csv_file_name)

    sample_variable_df.to_csv(id_ra_dec_csv_file_path, header=False, index=False)

    elapsed_time = time.time() - start_time
    print 'P3 Time: {} hour'.format(elapsed_time/3600)

    # ----------------------------------------------------------------------------------------

    start_time = time.time()

    with psycopg2.connect(psql_login) as connection:
        responses = inquire_bogus_transient(connection, id_ra_dec_csv_file_path, temp_table_name='hamasaki_source_sub')
        match_df = pd.DataFrame(responses, columns=['source_id1','source_id2'])
        connection.commit()

    elapsed_time = time.time() - start_time
    print 'P4 Time: {} hour'.format(elapsed_time/3600)

    # ----------------------------------------------------------------------------------------

    start_time = time.time()

    match_df = match_df[match_df['source_id1'] != match_df['source_id2']]
    group_ids = match_df.groupby('source_id1').agg(lambda x: set(x)).reset_index(inplace=False)
    match_df['group_id_list'] = match_df.apply(make_group_id, group_id=group_ids, axis=1)

    elapsed_time = time.time() - start_time
    print 'P5 Time: {} hour'.format(elapsed_time/3600)

    # ----------------------------------------------------------------------------------------

    bogus_transient_series = match_df.drop_duplicates(subset='group_id_list')['group_id_list']
    bogus_transient_list = [[int(source_id) for source_id in bogus_tran_str.split(',')] for bogus_tran_str in bogus_transient_series]

    file_name = 'bogus_transient_' + keywords + '.csv'
    output_file_path = os.path.join(output_path, file_name)

    with open(output_file_path, 'w') as bogus_transient_handle:
        writer = csv.writer(bogus_transient_handle, lineterminator='\n')
        writer.writerows(bogus_transient_list)



def inquire_bogus_transient(connection, id_ra_dec_csv_file_path, temp_table_name):
    
    with connection.cursor() as cur:
        query = 'CREATE TEMPORARY TABLE {} (source_id BIGINT, ra DOUBLE PRECISION, dec DOUBLE PRECISION)'.format(temp_table_name)
        cur.execute(query)
        # Insert (ra,dec)data to my_table
        with open(id_ra_dec_csv_file_path, mode='r') as file_handle:
            cur.copy_from(file_handle, temp_table_name, sep=',')

    with connection.cursor('match_my_table') as cur: # ---!!!CAUTION!!!--- Heavy Processing
        cur.itersize = 10000
        # Search ps1 catalog
        match_rad = 3.0/3600
        query = 'SELECT t."source_id", ss.* FROM {} AS t LEFT JOIN LATERAL (SELECT s."source_id" FROM {} AS s WHERE q3c_join(t.ra, t.dec, s.ra, s.dec, {}) ORDER BY q3c_dist(t.ra, t.dec, s.ra, s.dec) ASC) AS ss ON true'.format(temp_table_name, temp_table_name, match_rad)
        cur.execute(query)
        responses = cur.fetchall()

    return responses



def inquire_nearest_ps_star(connection, id_ra_dec_csv_file_path, temp_table_name, search_radius=15.0/3600):
    with connection.cursor() as cur:
        query = 'CREATE TEMPORARY TABLE {} (source_id BIGINT, ra DOUBLE PRECISION, dec DOUBLE PRECISION)'.format(temp_table_name)
        cur.execute(query)
        # Insert (ra,dec)data to my_table
        with open(id_ra_dec_csv_file_path, mode='r') as file_handle:
            cur.copy_from(file_handle, temp_table_name, sep=',')

    with connection.cursor('match_my_table') as cur: # ---!!!CAUTION!!!--- Heavy Processing
        cur.itersize = 10000

        # Search ps1 catalog
        query = 'SELECT ss.* FROM {} AS t LEFT JOIN LATERAL (SELECT s."psId", s."objInfoFlag", s."rmag", q3c_dist(t.ra, t.dec, s.ra, s.dec) FROM ps_bright AS s WHERE q3c_join(t.ra, t.dec, s.ra, s.dec, {}) ORDER BY q3c_dist(t.ra, t.dec, s.ra, s.dec) ASC LIMIT 1 ) AS ss ON true'.format(temp_table_name, search_radius)
        cur.execute(query)
        responses = cur.fetchall()

    data = [[0 if num is None else num for num in ps_param] for ps_param in responses]
    nearest_ps1_df = pd.DataFrame(data, columns=['psId','objInfoFlag','rmag','dist'])
    nearest_ps1_df['psType'] = nearest_ps1_df.apply(get_ps_data, axis=1)

    return nearest_ps1_df



def get_ps_data(match_series):
    pstype = None
    if match_series['objInfoFlag'] == match_series['objInfoFlag']:
        objInfoFlag = int(match_series['objInfoFlag'])
        pstype = 'star'
        if objInfoFlag:
            for iobj,objInfo in enumerate(['FEW','POOR','ICRF_QSO','HERN_QSO_P60','HERN_QSO_P05',
                                           'HERN_RRL_P60','HERN_RRL_P05','HERN_VARIABLE','TRANSIENT',
                                           'HAS_SOLSYS_DET','MOST_SOLSYS_DET','LARGE_PM','RAW_AVE',
                                           'FIT_AVE','FIT_PM','FIT_PAR','USE_AVE','USE_PM','USE_PAR',
                                           'NO_MEAN_ASTROM','STACK_FOR_MEAN','MEAN_FOR_STACK','BAD_PM',
                                           'EXT','EXT_ALT','GOOD','GOOD_ALT','GOOD_STACK','BEST_STACK',
                                           'SUSPECT_STACK','BAD_STACK']):
                if objInfoFlag & 2**iobj:
                    if objInfo == 'EXT': pstype = 'extended'
        else:
            pstype = None
    return pstype



def check_variable_star(pd_series):
    variable_star_flag = False
    if pd_series['psType'] == 'star':
        rmag = float(pd_series['rmag'])
        dist = float(pd_series['dist'])
        if (16 > rmag) and (10.0/3600 > dist): variable_star_flag = True
        if (16 <= rmag) and (3.0/3600 > dist): variable_star_flag = True
    return variable_star_flag



def make_group_id(pd_series, group_id):
    group_set = group_id[group_id['source_id1']==pd_series['source_id1']]['source_id2'].values[0]
    group_set.add(pd_series['source_id1'])
    return ','.join(map(str,group_set))



def random_coordinate(min_coordinate, max_coordinate, num):
    return (max_coordinate - min_coordinate) * np.random.rand(num) + min_coordinate



if __name__=='__main__':
    print 'Start.'
    main(date_min='2019-12-01', date_max='2020-02-01', target_num_of_sample_points=100000, output_path='./data')
    print 'Finish.'
