#! /usr/bin/env python

import tomoesnml
import numpy as np
import pandas as pd

import astropy.io.fits
import astropy.coordinates
import astropy.time
import astropy.units

import psycopg2
import warnings
import random
import os,sys
import shutil
import math
import time
import csv
from datetime import datetime
from pytz import timezone

import matplotlib.pyplot as plt



def main(output_directory, bogus_transient_source_id_csv_file_path, file_verbose_bool=True):

    ##---------- Initial setting

    psql_login = 'host=gwdata dbname=supernova user=tomoesn password='

    directory_in_path   = '/lustre/tomoesn/raw/'
    directory_temp_path = '/lustre/tomoesn/reference/'
    
    number_of_visits_threshold = 6

    warnings.simplefilter('ignore', astropy.wcs.wcs.FITSFixedWarning)

    random.seed(20200220)

    ##--------------------------------------- Read BOGUS Transient File
    
    with open(bogus_transient_source_id_csv_file_path,'r') as bogus_transient_handle:
        reader = csv.reader(bogus_transient_handle)
        bogus_transient_str_list = [source_ids for source_ids in reader]
        bogus_transient_list = [[int(source_id) for source_id in source_ids] for source_ids in bogus_transient_str_list]

    ##--------------------------------------- Filter BOGUS Transient with Number of Visits

    filtered_bogus_transient_list = [source_id_list for source_id_list in bogus_transient_list if len(source_id_list) < number_of_visits_threshold]
    number_of_source_id = len([source_id for source_id in source_id_list for source_id_list in filtered_bogus_transient_list])

    print '  Number of BOGUS Transient          : {}'.format(len(bogus_transient_list))
    print '  Number of Filtered BOGUS Transient : {}'.format(len(filtered_bogus_transient_list))
    print '  The time required for processing   : {:>5.1f} hours'.format(number_of_source_id*15.0/3600)
    print ''

    ##--------------------------------------- Remake 'tns_df' into unique data frame

    number_of_processed_files = 0
    number_of_successed_files = 0
    bogus_transient_number = 0

    for source_id_list in filtered_bogus_transient_list:
        bogus_transient_name = 'bogus_tran_' + str(bogus_transient_number)

        for source_id in source_id_list:
            with psycopg2.connect(psql_login) as connection:
                with connection.cursor() as cur: cur.execute('SET statement_timeout TO 60000') # 60 [sec]

                number_of_processed_files += 1
                start_time = time.time()

                source = inquire_source_data(connection, directory_in_path, source_id)
                # KEY: sourceId, rawId, detId, x, y, ra, dec, magauto, magautoerr, fwhm, a, b, theta,
                #      elongation, magmp, mjd, dateObs, fitsName, project, fitsPath

                idx1      = source['fitsPath'].rfind('/')
                idx2      = source['fitsPath'].rfind('.')
                keywords  = source['fitsPath'][idx1+1:idx2]
                fits_path = source['fitsPath']

                print '  BOGUS Transient Name : {}'.format(bogus_transient_name)
                print '  source ID : {}'.format(source['sourceId'])
                print '  file name : {}'.format(keywords)
                print '  file path : {}'.format(source['fitsPath'])
                print '  raw ID    : {}'.format(source['rawId'])

                # Check rawId
                fwhm, zero_mag, lim_mag, num_of_sources = inquire_analysis_stack_info(connection, source['rawId'])
                print '  fwhm, zero mag       : ' + str(fwhm) + ', ' + str(zero_mag)



                ###--------------------------------------- Main sequence

                ### Detect sources ###

                ##--------------------------------------- Detect sources

                new_sex = tomoesnml.SourceExtractor(fits_path,
                                                    detect_min_area  = 5,
                                                    detect_threshold = 1.4,
                                                    mag_zero_point   = 22.5,
                                                    pixel_scale      = 1.19,
                                                    seeing_fwhm      = 3.0)

                if new_sex.dataframe is None:
                    print ' [WARN] ' + keywords + ' : ' + 'No columns to parse from file !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                try:
                    table_name = 'hamasaki_table_' + str(os.getpid())
                    panstarrs_dataframe = get_panstarrs_data(connection, new_sex.dataframe,
                                                             temp_table_name=table_name,
                                                             max_num_of_inquire_objects=6000)
                except psycopg2.extensions.QueryCanceledError: # TimeoutError:
                    print ' [WARN] ' + keywords + ' : ' + 'Timeout error of inquire stars !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue
                except:
                    print ' [WARN] ' + keywords + ' : ' + 'Skip because occured exception !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                # panstarrs_dataframe.columns: ['PS_MATCH','PS_TYPE','PS_ID','OBJ_INFO_FLAG','PS_RMAG','DIST_PS1']
                new_sex.dataframe = new_sex.dataframe.join(panstarrs_dataframe) 

                ##--------------------------------------- Subtract fits image with embed ArtificialReals

                temp_remap_fits_path = keywords+'_temp_remap.fits'
                sub_fits_path        = keywords+'_sub.fits'

                flag, temp_fits_path = tomoesnml.imsub(connection, fits_path,
                                                       temp_remap_fits_name=temp_remap_fits_path,
                                                       subtracted_fits_name=sub_fits_path)
                if flag==1:
                    print ' [WARN] ' + keywords + ' : ' + 'Template image does not exist !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                ##--------------------------------------- Source detection from subtracted image

                sub_sex = tomoesnml.SourceExtractor(sub_fits_path,
                                                    detect_min_area  = 2,
                                                    detect_threshold = 3.,
                                                    mag_zero_point   = 20.0,
                                                    pixel_scale      = 1.19,
                                                    seeing_fwhm      = 4.0)

                with tomoesnml.FilterSource(temp_fits_path, temp_remap_fits_path, sub_fits_path) as filter_process:
                    sub_sex.dataframe['VALUE_ZERO_REF'] = sub_sex.dataframe.apply(filter_process.check_non_ref,    axis=1)
                    sub_sex.dataframe['SATURATION']     = sub_sex.dataframe.apply(filter_process.check_saturation, axis=1)
                    sub_sex.dataframe['MASK']           = sub_sex.dataframe.apply(filter_process.check_mask,       axis=1)
                    sub_sex.dataframe['EDGE']           = sub_sex.dataframe.apply(filter_process.check_edge,       axis=1)

                ##--------------------------------------- Output psf map ndarray and object data

                x0_img = source['x']
                y0_img = source['y']
                with tomoesnml.Cutout(temp_remap_fits_path, fits_path, sub_fits_path) as fits_handle:
                    ref_psf_nd = tomoesnml.Cutout.cutout_image(x0_img, y0_img, fits_handle.ref_hdu[0].data, psf_size=14)
                    new_psf_nd = tomoesnml.Cutout.cutout_image(x0_img, y0_img, fits_handle.new_hdu[0].data, psf_size=14)
                    sub_psf_nd = tomoesnml.Cutout.cutout_image(x0_img, y0_img, fits_handle.sub_hdu[0].data, psf_size=14)

                if sub_psf_nd.shape[0] != 29 or sub_psf_nd.shape[1] != 29 :
                    print '[INFO]  SKIP. (Mismatched Size)'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                source_psf_nd = np.stack([ref_psf_nd, new_psf_nd, sub_psf_nd], axis=-1)

                match_new_obj_df = new_sex.dataframe.query('(X_IMAGE - @x0_img)**2 + (Y_IMAGE - @y0_img)**2 <= (2.521)**2').iloc[:1,:]
                match_sub_obj_df = sub_sex.dataframe.query('(X_IMAGE - @x0_img)**2 + (Y_IMAGE - @y0_img)**2 <= (2.521)**2').iloc[:1,:]

                # make region
                reg_name = '{}.{}.{}.target.reg'.format(keywords, bogus_transient_name, source['sourceId'])
                tomoesnml.MakeReg(reg_name,
                                  np.array([source['ra']]),
                                  np.array([source['dec']]),
                                  numbers=None,
                                  values=None,
                                  color='pink',
                                  coordsys='FK5',
                                  y_desc=8.0/3600,
                                  circle_radius=3.0/3600)

                if match_new_obj_df.shape[0] == 0:
                    #print '------ Not match. new'
                    match_new_obj_df = pd.DataFrame(np.full((1, match_new_obj_df.shape[1]), 0), columns=match_new_obj_df.columns)
                if match_sub_obj_df.shape[0] == 0:
                    #print '------ Not match. sub'
                    match_sub_obj_df = pd.DataFrame(np.full((1, match_sub_obj_df.shape[1]), 0), columns=match_sub_obj_df.columns)

                shutil.copy(fits_path, keywords+'_new.fits')
                shutil.copy(temp_fits_path, keywords+'_temp.fits')
                source['bogus_trnasient_name'] = bogus_transient_name
                tomoe_df = pd.DataFrame([source])

                psf_file_name           = bogus_transient_name + '.' + str(source['sourceId']) + '.psf.npy'
                tomoe_param_file_name   = bogus_transient_name + '.' + str(source['sourceId']) + '.tomoe_params.csv'
                new_sex_param_file_name = bogus_transient_name + '.' + str(source['sourceId']) + '.new_sex_params.csv'
                sub_sex_param_file_name = bogus_transient_name + '.' + str(source['sourceId']) + '.sub_sex_params.csv'

                np.save(psf_file_name, source_psf_nd)
                tomoe_df.to_csv(tomoe_param_file_name, index=False)
                match_new_obj_df.to_csv(new_sex_param_file_name, index=False)
                match_sub_obj_df.to_csv(sub_sex_param_file_name, index=False)

                ##--------------------------------------- Move files and Remove file
                directory_out_path1 = os.path.join(output_directory, bogus_transient_name, 'analysis_data', str(source_id))
                if not os.path.isdir(directory_out_path1): os.makedirs(directory_out_path1)

                tomoesnml.move_files(psf_file_name           , directory_out_path1)
                tomoesnml.move_files(new_sex_param_file_name , directory_out_path1)
                tomoesnml.move_files(sub_sex_param_file_name , directory_out_path1)
                tomoesnml.move_files(tomoe_param_file_name   , directory_out_path1)

                if file_verbose_bool == True :
                    directory_out_path2 = os.path.join(output_directory, bogus_transient_name, 'all_data', str(source_id))
                    if not os.path.isdir(directory_out_path2): os.makedirs(directory_out_path2)
                    tomoesnml.move_files(keywords+'*', directory_out_path2)
                else :
                    tomoesnml.remove_files(keywords+'*')

                ##--------------------------------------- Output processing time
                elapsed_time = time.time() - start_time

                print '  Processed  number   : {} (included failure {} )'.format(number_of_successed_files+1, number_of_processed_files)
                print '  Processing time     : {}'.format(elapsed_time)
                print ''
                number_of_successed_files += 1
                connection.commit()

        bogus_transient_number += 1



def inquire_source_data(connection, directory_in_path, source_id):
    with connection.cursor() as cur:
        query = 'SELECT source_sub."rawId", source_sub."detId", source_sub."ra", source_sub."dec", source_sub."magauto", source_sub."magautoerr", source_sub."fwhm", source_sub."a", source_sub."b", source_sub."theta", source_sub."elongation", source_sub."magmp", raw."mjd", raw."dateObs", raw."fitsName", raw."project", source_sub."x", source_sub."y" FROM source_sub INNER JOIN raw ON source_sub."rawId" = raw."rawId" WHERE source_sub."sourceId" = {}'.format(source_id)
        cur.execute(query)
        res = cur.fetchone()

    source = {}
    source['sourceId']   = source_id
    source['rawId']      = res[0]
    source['detId']      = res[1]
    source['ra']         = res[2]
    source['dec']        = res[3]
    source['magauto']    = res[4]
    source['magautoerr'] = res[5]
    source['fwhm']       = res[6]
    source['a']          = res[7]
    source['b']          = res[8]
    source['theta']      = res[9]
    source['elongation'] = res[10]
    source['magmp']      = res[11]
    source['mjd']        = res[12]
    source['dateObs']    = res[13]
    source['fitsName']   = res[14]
    source['project']    = res[15]
    source['x']          = res[16]
    source['y']          = res[17]
    source['fitsPath']   = os.path.join(directory_in_path + source['fitsName'])
    return source



def inquire_analysis_stack_info(connection, raw_id):
    with connection.cursor('inquire_tns_data') as cur:
        query = 'SELECT "fwhm","zeromag","limmag","numofsources","sn5mag","sn5magsub" FROM analysis_stack WHERE "rawId" = \'{}\' '.format(raw_id)
        cur.execute(query)
        row = cur.fetchone()
        if(row != None):
            fwhm           = row[0]
            zero_mag       = row[1]
            lim_mag        = row[2]
            num_of_sources = row[3]
            sn5_mag        = row[4]
            sn5_mag_sub    = row[5]

            return fwhm, zero_mag, lim_mag, num_of_sources
        else:
            return 0, 0, 0, 0



def get_panstarrs_data(connection, dataframe, temp_table_name='hamasaki_table', max_num_of_inquire_objects=8000):

    def inquire_psql_database(ra_np, dec_np, temp_table_name):
        ra_dec_np       = np.stack([ra_np,dec_np], axis=-1)
        ra_dec_tuple    = tuple(map(tuple,ra_dec_np)) # ra_dec_tuple = ((ra1, dec1),(ra2, dec2)...(ra, dec))
        ra_dec_str      = ','.join(map(str, ra_dec_tuple))
        with connection.cursor() as cur:
            query = 'CREATE TEMPORARY TABLE {} (ra DOUBLE PRECISION, dec DOUBLE PRECISION)'.format(temp_table_name)
            cur.execute(query)

            # Insert (ra,dec)data to my_table
            query = 'INSERT INTO {} (ra,dec) VALUES {}'.format(temp_table_name, ra_dec_str)
            cur.execute(query)
        with connection.cursor('match_ps_table') as cur: # ---!!!CAUTION!!!--- Heavy Processing
            cur.itersize = 10000
            # Search ps1 catalog
            query = 'SELECT ss.* FROM {} AS t LEFT JOIN LATERAL (SELECT s."psId", s."objInfoFlag", s."rmag", q3c_dist(t.ra, t.dec, s.ra, s.dec) FROM ps_bright AS s WHERE q3c_join(t.ra, t.dec, s.ra, s.dec, 1./60) ORDER BY q3c_dist(t.ra, t.dec, s.ra, s.dec) ASC LIMIT 1 ) AS ss ON true'.format(temp_table_name)
            cur.execute(query)
            responses = cur.fetchall()
            data = [[0 if num is None else num for num in ps_param] for ps_param in responses]
        return data

    def get_ps_data(ps_info):
        match_rad   = 3./3600 # [arcsec]
        match       = False
        pstype      = 'none'
        if ps_info:
            psId, objInfoFlag, rmag, dist_ps1 = ps_info
            pstype = 'star'

            if objInfoFlag:
                for iobj,objInfo in enumerate(['FEW','POOR','ICRF_QSO','HERN_QSO_P60','HERN_QSO_P05',
                                               'HERN_RRL_P60','HERN_RRL_P05','HERN_VARIABLE','TRANSIENT',
                                               'HAS_SOLSYS_DET','MOST_SOLSYS_DET','LARGE_PM','RAW_AVE',
                                               'FIT_AVE','FIT_PM','FIT_PAR','USE_AVE','USE_PM','USE_PAR',
                                               'NO_MEAN_ASTROM','STACK_FOR_MEAN','MEAN_FOR_STACK','BAD_PM',
                                               'EXT','EXT_ALT','GOOD','GOOD_ALT','GOOD_STACK','BEST_STACK',
                                               'SUSPECT_STACK','BAD_STACK']):
                    if objInfoFlag & 2**iobj:
                        if objInfo == 'EXT': pstype = 'extended'

        if dist_ps1:
            if dist_ps1 <= match_rad: match = True

        return [match, pstype, psId, objInfoFlag, rmag, dist_ps1]

    ra_np  = dataframe['ALPHA_J2000'].values
    dec_np = dataframe['DELTA_J2000'].values
    if ra_np.shape[0] > max_num_of_inquire_objects: raise Exception('Too many objects : {}'.format(ra_np.shape[0]))
    try: responses = inquire_psql_database(ra_np, dec_np, temp_table_name)
    except: raise
    else:
        ps_list = [get_ps_data(res) for res in responses]
        ps_df = pd.DataFrame(ps_list, columns=['PS_MATCH','PS_TYPE','PS_ID','OBJ_INFO_FLAG','PS_RMAG','DIST_PS1'])
        return ps_df



if __name__=='__main__':
    print 'Start.'
    output_directory = '/gwkiso/hamasaki/test_data/bogus_transient'
    csv_file_path = 'bogus_transient_2019-12-01_2020-02-01.csv'
    main(output_directory, csv_file_path, file_verbose_bool=False)
    print 'Finish.'
    exit(0)
