#! /usr/bin/env python

import tomoesnml
import numpy as np
import pandas as pd

import astropy.io.fits
import astropy.coordinates
import astropy.time
import astropy.units

import psycopg2
import warnings
import random
import os,sys
import shutil
import math
import time
import csv
from datetime import datetime
from pytz import timezone

import matplotlib.pyplot as plt



def main(date_min, date_max, output_directory, file_verbose_bool=True):

    ##---------- Initial setting

    psql_login = 'host=gwdata dbname=supernova user=tomoesn password='

    directory_in_path   = '/lustre/tomoesn/raw/'
    directory_temp_path = '/lustre/tomoesn/reference/'

    warnings.simplefilter('ignore', astropy.wcs.wcs.FITSFixedWarning)

    random.seed(20200220)

    ##--------------------------------------- Get data from transient name server
    print "  Get transient data from TNS ..."
    tns_df = get_tns_data(psql_login, date_min, date_max)
    if tns_df is None: sys.exit('No TNS data.')
    # tns_df.columns : ['TNS_name','TNS_ra','TNS_dec','TNS_aliasName','TNS_discDate','TNS_discMag','TNS_discFilter','TNS_tnsId','TNS_mjd']

    ##--------------------------------------- Print infos
    print '  Number of TNS objects        : {}'.format(tns_df.shape[0])
    print '  Number of TNS unique objects : {}'.format(tns_df['TNS_tnsId'].unique().shape[0])
    print ''

    ##--------------------------------------- Remake 'tns_df' into unique data frame
    tns_df.drop_duplicates(keep='last',subset='TNS_tnsId',inplace=True)



    number_of_processed_files = 0
    number_of_successed_files = 0
    number_of_detected_tomoe  = 0

    for tns_index, tns_data_series in tns_df.iterrows():

        number_of_processed_files += 1

        today_mjd = astropy.time.Time(datetime.now(timezone('UTC')).replace(tzinfo=None), scale='utc').mjd

        ##--------------------------------------- Get tomoe detected data
        tomoe_df = get_tomoe_detected_data(psql_login, tns_data_series)
        # tomoe_df.columns : ['raw_id', 'det_id', 'source_id', 'x_image', 'y_image', 'ra', 'dec', 'mag_auto', 'fwhm',
        #                     'cnn_cand', 'param_cand', 'variable_id', 'transient_id', 'cnn_prob', 'mjd', 'fits_name', 'dist']

        if tomoe_df is None:
            print '[INFO] No tomoe detected data. [ TNS name : {} ]'.format(tns_data_series['TNS_name'])
            continue

        tomoe_df['fits_name'] = directory_in_path + tomoe_df['fits_name']
        number_of_detected_tomoe += 1

        for tomoe_index, tomoe_data_series in tomoe_df.iterrows():
            with psycopg2.connect(psql_login) as connection:
                with connection.cursor() as cur: cur.execute('SET statement_timeout TO 60000') # 60 [sec]
                start_time = time.time()

                det_id    = int(tomoe_data_series['det_id'])
                raw_id    = int(tomoe_data_series['raw_id'])
                fits_path = tomoe_data_series['fits_name']
                idx1      = fits_path.rfind('/')
                idx2      = fits_path.rfind('.')
                keywords  = fits_path[idx1+1:idx2]

                print '  TNS name : {}'.format(tns_data_series['TNS_name'])
                print '  TNS mjd  : {}'.format(tns_data_series['TNS_mjd'])
                print '  Today    : {}'.format(today_mjd)
                print '  Today - TNS mjd : {}'.format(today_mjd - tns_data_series['TNS_mjd'])
                print '  File name : {}'.format(keywords)
                print '  File path : {}'.format(fits_path)
                print '  raw id    : {}'.format(raw_id)

                #if today_mjd - tomoe_data_series['mjd'] > 90 : # Over passed 3 month
                #    print '[INFO] SKIP. [ERR_01]'
                #    print ''
                #    connection.commit()
                #    continue

                if tomoe_data_series['mjd'] < tns_data_series['TNS_mjd']:
                    print '[INFO] SKIP. [ERR_02]'
                    print ''
                    connection.commit()
                    continue

                # Check rawId
                fwhm, zero_mag, lim_mag, num_of_sources = inquire_analysis_stack_info(connection, raw_id)
                print '  fwhm, zero mag      : ' + str(fwhm) + ', ' + str(zero_mag)

                ###--------------------------------------- Main sequence

                ### Detect sources ###

                ##--------------------------------------- Detect sources

                new_sex = tomoesnml.SourceExtractor(fits_path,
                                                    detect_min_area  = 5,
                                                    detect_threshold = 1.4,
                                                    mag_zero_point   = 22.5,
                                                    pixel_scale      = 1.19,
                                                    seeing_fwhm      = 3.0)

                if new_sex.dataframe is None:
                    print ' [WARN] ' + keywords + ' : ' + 'No columns to parse from file !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                try:
                    table_name = 'hamasaki_table_' + str(os.getpid())
                    panstarrs_dataframe = get_panstarrs_data(connection, new_sex.dataframe,
                                                             temp_table_name=table_name,
                                                             max_num_of_inquire_objects=6000)
                except psycopg2.extensions.QueryCanceledError: # TimeoutError:
                    print ' [WARN] ' + keywords + ' : ' + 'Timeout error of inquire stars !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue
                except:
                    print ' [WARN] ' + keywords + ' : ' + 'Skip because occured exception !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                # panstarrs_dataframe.columns: ['PS_MATCH','PS_TYPE','PS_ID','OBJ_INFO_FLAG','PS_RMAG','DIST_PS1']
                new_sex.dataframe = new_sex.dataframe.join(panstarrs_dataframe)

                ##--------------------------------------- Subtract fits image with embed ArtificialReals

                temp_remap_fits_path = keywords+'_temp_remap.fits'
                sub_fits_path        = keywords+'_sub.fits'

                flag, temp_fits_path = tomoesnml.imsub(connection, fits_path,
                                                       temp_remap_fits_name=temp_remap_fits_path,
                                                       subtracted_fits_name=sub_fits_path)
                if flag==1:
                    print ' [WARN] ' + keywords + ' : ' + 'Template image does not exist !'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                ##--------------------------------------- Source detection from subtracted image

                sub_sex = tomoesnml.SourceExtractor(sub_fits_path,
                                                    detect_min_area  = 2,
                                                    detect_threshold = 3.,
                                                    mag_zero_point   = 20.0,
                                                    pixel_scale      = 1.19,
                                                    seeing_fwhm      = 4.0)

                with tomoesnml.FilterSource(temp_fits_path, temp_remap_fits_path, sub_fits_path) as filter_process:
                    sub_sex.dataframe['VALUE_ZERO_REF'] = sub_sex.dataframe.apply(filter_process.check_non_ref,    axis=1)
                    sub_sex.dataframe['SATURATION']     = sub_sex.dataframe.apply(filter_process.check_saturation, axis=1)
                    sub_sex.dataframe['MASK']           = sub_sex.dataframe.apply(filter_process.check_mask,       axis=1)
                    sub_sex.dataframe['EDGE']           = sub_sex.dataframe.apply(filter_process.check_edge,       axis=1)

                ##--------------------------------------- Output psf map ndarray and object data

                x0_img = tomoe_data_series['x_image']
                y0_img = tomoe_data_series['y_image']
                with tomoesnml.Cutout(temp_remap_fits_path, fits_path, sub_fits_path) as fits_handle:
                    ref_psf_nd = tomoesnml.Cutout.cutout_image(x0_img, y0_img, fits_handle.ref_hdu[0].data, psf_size=14)
                    new_psf_nd = tomoesnml.Cutout.cutout_image(x0_img, y0_img, fits_handle.new_hdu[0].data, psf_size=14)
                    sub_psf_nd = tomoesnml.Cutout.cutout_image(x0_img, y0_img, fits_handle.sub_hdu[0].data, psf_size=14)

                if sub_psf_nd.shape[0] != 29 or sub_psf_nd.shape[1] != 29 :
                    print '[INFO]  SKIP. (Mismatched Size)'
                    print ''
                    tomoesnml.remove_files(keywords+'*')
                    connection.commit()
                    continue

                source_psf_nd = np.stack([ref_psf_nd, new_psf_nd, sub_psf_nd], axis=-1)

                match_new_obj_df = new_sex.dataframe.query('(X_IMAGE - @x0_img)**2 + (Y_IMAGE - @y0_img)**2 <= (2.521)**2').iloc[:1,:]
                match_sub_obj_df = sub_sex.dataframe.query('(X_IMAGE - @x0_img)**2 + (Y_IMAGE - @y0_img)**2 <= (2.521)**2').iloc[:1,:]
                


                # make region
                reg_name = keywords + '.' + str(tns_data_series['TNS_tnsId']) + '.target.reg'                
                tomoesnml.MakeReg(reg_name,
                                  np.array([tomoe_data_series.values[3]]),
                                  np.array([tomoe_data_series.values[4]]),
                                  values=np.array([tomoe_data_series.values[7]]),
                                  color='lime')



                if match_new_obj_df.shape[0] == 0:
                    #print '------ Not match. new'
                    match_new_obj_df = pd.DataFrame(np.full((1, match_new_obj_df.shape[1]), 0), columns=match_new_obj_df.columns)
                if match_sub_obj_df.shape[0] == 0:
                    #print '------ Not match. sub'
                    match_sub_obj_df = pd.DataFrame(np.full((1, match_sub_obj_df.shape[1]), 0), columns=match_sub_obj_df.columns)

                shutil.copy(fits_path, keywords+'_new.fits')

                #print '  DET {} (PSF : {}, TNS : {}, NEW : {}, SUB : {})'.format(det_id, source_psf_nd.shape, tns_data_series.values.shape,
                #                                                                 match_new_obj_df.values.shape, match_sub_obj_df.values.shape)

                psf_file_name           = str(tns_data_series['TNS_tnsId']) + '.' + str(tomoe_data_series['source_id']) + '.psf.npy'
                tns_param_file_name     = str(tns_data_series['TNS_tnsId']) + '.' + str(tomoe_data_series['source_id']) + '.tns_params.csv'
                tomoe_param_file_name   = str(tns_data_series['TNS_tnsId']) + '.' + str(tomoe_data_series['source_id']) + '.tomoe_params.csv'
                new_sex_param_file_name = str(tns_data_series['TNS_tnsId']) + '.' + str(tomoe_data_series['source_id']) + '.new_sex_params.csv'
                sub_sex_param_file_name = str(tns_data_series['TNS_tnsId']) + '.' + str(tomoe_data_series['source_id']) + '.sub_sex_params.csv'

                np.save(psf_file_name, source_psf_nd)
                pd.DataFrame([tns_data_series]).to_csv(tns_param_file_name, index=False)
                pd.DataFrame([tomoe_data_series]).to_csv(tomoe_param_file_name, index=False)
                match_new_obj_df.to_csv(new_sex_param_file_name, index=False)
                match_sub_obj_df.to_csv(sub_sex_param_file_name, index=False)

                ##--------------------------------------- Move files and Remove file
                det_dir_name = 'Q' + str(int(tomoe_data_series['det_id']))
                directory_out_path1 = os.path.join(output_directory, det_dir_name, 'analysis_data', str(tns_data_series['TNS_tnsId']))
                if not os.path.isdir(directory_out_path1): os.makedirs(directory_out_path1)
                tomoesnml.move_files(psf_file_name           , directory_out_path1)
                tomoesnml.move_files(tns_param_file_name     , directory_out_path1)
                tomoesnml.move_files(tomoe_param_file_name   , directory_out_path1)
                tomoesnml.move_files(new_sex_param_file_name , directory_out_path1)
                tomoesnml.move_files(sub_sex_param_file_name , directory_out_path1)

                if file_verbose_bool == True :
                    directory_out_path2 = os.path.join(output_directory, det_dir_name, 'all_data', keywords)
                    if not os.path.isdir(directory_out_path2): os.makedirs(directory_out_path2)
                    tomoesnml.move_files(keywords+'*', directory_out_path2)
                else :
                    tomoesnml.remove_files(keywords+'*')

                ##--------------------------------------- Output processing time
                elapsed_time = time.time() - start_time

                print '  Processed  number   : {} (included failure {} )'.format(number_of_successed_files+1, number_of_processed_files)
                print '  Processing time     : {}'.format(elapsed_time)
                print ''
                number_of_successed_files += 1
                connection.commit()

    print '  Number of processed files         : {} (included failure {} )'.format(number_of_successed_files, number_of_processed_files)
    print '  Number of TNS unique objects      : {}'.format(tns_df['TNS_tnsId'].unique().shape[0])
    print '  Number of Tomo-e detected objects : {}'.format(number_of_detected_tomoe)



def get_panstarrs_data(connection, dataframe, temp_table_name='hamasaki_table', max_num_of_inquire_objects=8000):

    def inquire_psql_database(ra_np, dec_np, temp_table_name):
        ra_dec_np       = np.stack([ra_np,dec_np], axis=-1)
        ra_dec_tuple    = tuple(map(tuple,ra_dec_np)) # ra_dec_tuple = ((ra1, dec1),(ra2, dec2)...(ra, dec))
        ra_dec_str      = ','.join(map(str, ra_dec_tuple))
        with connection.cursor() as cur:
            query = 'CREATE TEMPORARY TABLE {} (ra DOUBLE PRECISION, dec DOUBLE PRECISION)'.format(temp_table_name)
            cur.execute(query)

            # Insert (ra,dec)data to my_table
            query = 'INSERT INTO {} (ra,dec) VALUES {}'.format(temp_table_name, ra_dec_str)
            cur.execute(query)
        with connection.cursor('match_ps_table') as cur: # ---!!!CAUTION!!!--- Heavy Processing
            cur.itersize = 10000
            # Search ps1 catalog
            query = 'SELECT ss.* FROM {} AS t LEFT JOIN LATERAL (SELECT s."psId", s."objInfoFlag", s."rmag", q3c_dist(t.ra, t.dec, s.ra, s.dec) FROM ps_bright AS s WHERE q3c_join(t.ra, t.dec, s.ra, s.dec, 1./60) ORDER BY q3c_dist(t.ra, t.dec, s.ra, s.dec) ASC LIMIT 1 ) AS ss ON true'.format(temp_table_name)
            cur.execute(query)
            responses = cur.fetchall()
            data = [[0 if num is None else num for num in ps_param] for ps_param in responses]
        return data

    def get_ps_data(ps_info):
        match_rad   = 3./3600 # [arcsec]
        match       = False
        pstype      = None
        if ps_info:
            psId, objInfoFlag, rmag, dist_ps1 = ps_info
            pstype = 'star'

            if objInfoFlag:
                for iobj,objInfo in enumerate(['FEW','POOR','ICRF_QSO','HERN_QSO_P60','HERN_QSO_P05',
                                               'HERN_RRL_P60','HERN_RRL_P05','HERN_VARIABLE','TRANSIENT',
                                               'HAS_SOLSYS_DET','MOST_SOLSYS_DET','LARGE_PM','RAW_AVE',
                                               'FIT_AVE','FIT_PM','FIT_PAR','USE_AVE','USE_PM','USE_PAR',
                                               'NO_MEAN_ASTROM','STACK_FOR_MEAN','MEAN_FOR_STACK','BAD_PM',
                                               'EXT','EXT_ALT','GOOD','GOOD_ALT','GOOD_STACK','BEST_STACK',
                                               'SUSPECT_STACK','BAD_STACK']):
                    if objInfoFlag & 2**iobj:
                        if objInfo == 'EXT': pstype = 'extended'

        if dist_ps1:
            if dist_ps1 <= match_rad: match = True        

        return [match, pstype, psId, objInfoFlag, rmag, dist_ps1]

    ra_np  = dataframe['ALPHA_J2000'].values
    dec_np = dataframe['DELTA_J2000'].values
    if ra_np.shape[0] > max_num_of_inquire_objects: raise Exception('Too many objects : {}'.format(ra_np.shape[0]))
    try: responses = inquire_psql_database(ra_np, dec_np, temp_table_name)
    except: raise
    else:
        ps_list = [get_ps_data(res) for res in responses]
        ps_df = pd.DataFrame(ps_list, columns=['PS_MATCH','PS_TYPE','PS_ID','OBJ_INFO_FLAG','PS_RMAG','DIST_PS1'])
        return ps_df

def get_tns_data(psql_login, date_min, date_max):
    with psycopg2.connect(psql_login) as connection:
        with connection.cursor('inquire_tns_data') as cur:
            cur.itersize = 3000
            query = 'SELECT "name","ra","dec","aliasName","discDate","discMag","discFilter","tnsId" FROM tns WHERE "discDate">=\'{}\' AND "discDate"<=\'{}\' '.format(date_min,date_max)
            #query = 'SELECT "name","ra","dec","aliasName","discDate","discMag","discFilter","tnsId" FROM tns WHERE "tnsId"=46815' # for test
            cur.execute(query)
            responses = cur.fetchall()
        connection.commit()
    if(responses != None):
        tns_df = pd.DataFrame(responses, columns=['TNS_name','TNS_ra','TNS_dec','TNS_aliasName','TNS_discDate','TNS_discMag','TNS_discFilter','TNS_tnsId'])
        tns_df['TNS_mjd'] = tns_df.apply(convert_mjd, axis=1)
        return tns_df
    else:
        return None

def convert_mjd(tns_data_series):
    date_time_astropy = astropy.time.Time(str(tns_data_series['TNS_discDate']), format='iso', scale='utc')
    return date_time_astropy.mjd

def get_tomoe_detected_data(psql_login, tns_data_series):
    with psycopg2.connect(psql_login) as connection:
        responses   = inquire_tomoe_detected_data(connection, tns_data_series['TNS_ra'], tns_data_series['TNS_dec'], tns_data_series['TNS_mjd'])
        cols        = ['raw_id', 'det_id', 'source_id', 'x_image', 'y_image', 'ra', 'dec', 'mag_auto', 'fwhm',
                       'cnn_cand', 'param_cand', 'variable_id', 'transient_id', 'cnn_prob', 'mjd', 'fits_name', 'dist']
        connection.commit()
    if responses:
        tomoe_df = pd.DataFrame(responses, columns=cols)
    else:
        tomoe_df = None # pd.DataFrame(np.full((1, len(cols)), None), columns=cols)
    return tomoe_df

def inquire_tomoe_detected_data(connection, ra, dec, mjd):
    #print ra, dec, mjd
    with connection.cursor('inquire_tns_data') as cur:
        cur.itersize = 3000
        query = 'SELECT source_sub."rawId", source_sub."detId", source_sub."sourceId", source_sub."x", source_sub."y", source_sub."ra", source_sub."dec", source_sub."magauto", source_sub."fwhm", source_sub."cnncand", source_sub."paramcand", source_sub."variableId", transient."transientId", source_sub."cnnprob", raw."mjd", raw."fitsName", q3c_dist(source_sub."ra", source_sub."dec", {}, {}) AS dist FROM source_sub INNER JOIN raw ON source_sub."rawId" = raw."rawId" LEFT JOIN transient ON source_sub."variableId" = transient."variableId" WHERE q3c_radial_query(source_sub."ra", source_sub."dec", {}, {}, {}) AND raw."mjd" > {} ORDER BY dist ASC'.format(ra, dec, ra, dec, 3./3600., mjd)
        cur.execute(query)
        res = cur.fetchall()
        data = [[0 if num is None else num for num in param] for param in res]
    return data

def inquire_analysis_stack_info(connection, raw_id):
    with connection.cursor('inquire_tns_data') as cur:
        query = 'SELECT "fwhm","zeromag","limmag","numofsources","sn5mag","sn5magsub" FROM analysis_stack WHERE "rawId" = \'{}\' '.format(raw_id)
        cur.execute(query)
        row = cur.fetchone()
        if(row != None):
            fwhm           = row[0]
            zero_mag       = row[1]
            lim_mag        = row[2]
            num_of_sources = row[3]
            sn5_mag        = row[4]
            sn5_mag_sub    = row[5]

            return fwhm, zero_mag, lim_mag, num_of_sources
        else:
            return 0, 0, 0, 0



if __name__=='__main__':
    print 'Start.'
    output_directory = '/gwkiso/hamasaki/test_data/real_transient'
    main(date_min='2019-12-01', date_max='2020-02-01', output_directory=output_directory, file_verbose_bool=True)
    print 'Finish.'
    exit(0)
