import astropy.io.fits
import astropy.wcs
import pandas as pd
import numpy as np
import psycopg2
import random
import os,sys
import glob
import math
import time
from timeout_decorator import timeout, TimeoutError

# Module list ============================================
#  class: SourceExtractor
#  class: MakeReg
#  class: NormedPSF
#  class: ArtificialReal
#  class: FilterSource
#  class: Cutout
#  func : remove_files
#  func : imsub          (Require: psycopg2.connect)
#  func : template_image <<< Use in imsub
# ========================================================



class SourceExtractor(object):
    """
    For manage SExtractor.

    Attributes
    ----------
    detect_min_area : int or float, default 5
    detect_threshold : int or float, default 1.8
    mag_zero_point : int or float, default 22.5
    pixel_scale : int or float, default 1.19
    seeing_fwhm : int or float, default 3.0
        see SExtractor manual.

    fits_path : str
        fits file path. (e.g.: ./test/python.fits)
    fits_name : str
        fits file name. (e.g.: python.fits)
    cat_name : str
        cat file name. (e.g.: python.cat)

    Examples
    --------
    >>> sex_fits = SourceExtractor('./python.fits')
    >>> sex_fits.dataframe # Pandas Dataframe of detected sources

    """
    def __init__(self, file_path, cat_file_stem=None,
                 detect_min_area=5,
                 detect_threshold=1.8,
                 mag_zero_point=22.5,
                 pixel_scale=1.19,
                 seeing_fwhm=3.0):
        """
        For running SExtractor.

        Parameters
        ----------
        file_path : str
            fits file path.
        cat_file_stem : str, default None
            for output stem name of cat file.

        detect_min_area : int or float, default 5
        detect_threshold : int or float, default 1.8
        mag_zero_point : int or float, default 22.5
        pixel_scale : int or float, default 1.19
        seeing_fwhm : int or float, default 3.0
            see SExtractor manual.
        
        """
        self.fits_path  = file_path # file path
        self.fits_name  = os.path.basename(self.fits_path) # file name
        if cat_file_stem is None:
            self.cat_name = self.fits_name.replace('.fits','.cat') # cat file name
        else:
            self.cat_name = cat_file_stem + '.cat' # cat file name

        self.detect_min_area  = detect_min_area
        self.detect_threshold = detect_threshold
        self.mag_zero_point   = mag_zero_point
        self.pixel_scale      = pixel_scale
        self.seeing_fwhm      = seeing_fwhm

        self.execute() # Execute SExtractor
        self.make_dataframe() # Make Pandas Dataframe of detected sources.
    
    def execute(self):
        _command_sex = ('sex ' + str(self.fits_path) + ' -CATALOG_NAME ' + self.cat_name +
                        ' -DETECT_MINAREA ' + str(self.detect_min_area) + ' -DETECT_THRESH ' + str(self.detect_threshold) +
                        ' -MAG_ZEROPOINT ' + str(self.mag_zero_point) + ' -PIXEL_SCALE ' + str(self.pixel_scale) +
                        ' -SEEING_FWHM ' + str(self.seeing_fwhm) + '   >/dev/null 2>&1')
        #print _command_sex
        os.system(_command_sex)

    def make_dataframe(self):
        try:
            # read data
            source_df = pd.read_csv(self.cat_name,delim_whitespace=True,comment='#',header=None, engine='python')

            # read header
            with open(self.cat_name,'r') as f_cat:
                col_num = source_df.shape[1]
                key = []

                i = 0
                for line in f_cat.readlines():
                    if line.startswith('#'):
                        legend = line.split()
                        key.append(legend[2])
                        i += 1
                    if i >= col_num :
                        break

        except pd.errors.EmptyDataError:
            self.dataframe = None

        else:        
            source_df.columns = key
            self.dataframe = source_df



class MakeReg(object):
    """
    Make reg file for ds9.

    Attributes
    ----------
    reg_path : str
    x : numpy.ndarray or list, shape (n,)
    y : numpy.ndarray or list, shape (n,)
    numbers : numpy.ndarray or list, shape (n,)
    values : numpy.ndarray or list, shape (n,)
    color : str, default green
    coordinate_system : str, default IMAGE
        Coordinate system.
    y_desc : int, default 20
        Description height from center of a circle.
    circle_radius : int, default 15

    Examples
    --------
    >>> MakeReg('./python.reg', x, y, numbers=num)
    """
    def __init__(self, file_path,
                 x, y,
                 numbers=None,
                 values=None,
                 color='green',
                 coordsys='IMAGE',
                 y_desc=20,
                 circle_radius=15):
        """
        Parameters
        ----------
        file_path : str
            reg file path.
        x : numpy.ndarray or list, shape (n,)
        y : numpy.ndarray or list, shape (n,)
        numbers : numpy.ndarray or list, shape (n,)
        values : numpy.ndarray or list, shape (n,)
        color : str, default green
        coordsys : str, default IMAGE
            Coordinate system.
        y_desc : int, default 20
            Description height from center of a circle.
        circle_radius : int, default 15
        """
        self.reg_path = file_path # file path
        self.x = x
        self.y = y
        if numbers is not None: self.numbers = numbers
        else: self.numbers = np.full(x.shape[0], None)
        if values is not None:  self.values  = values
        else: self.values  = np.full(x.shape[0], None)
        self.color = color
        self.coordinate_system = coordsys
        self.y_desc = y_desc
        self.circle_radius = circle_radius

        self.output()

    def __repr__(self):
        return ''

    def output(self):
        with open(self.reg_path, 'w') as reg_handle:
            reg_handle.write(self.coordinate_system + ' \n')
            for x, y, number, value in zip(self.x, self.y, self.numbers, self.values):
                reg1, reg2 = MakeReg.region_syntax(x, y, number, value,
                                                   circle_radius=self.circle_radius,
                                                   circle_color=self.color,
                                                   y_desc=self.y_desc)
                reg_handle.write(reg1)
                reg_handle.write(reg2)

    @staticmethod
    def region_syntax(x, y, number=None, value=None, circle_radius=15, circle_color='green', y_desc=20):
        reg_syntax1 = ('circle( '+
                       str(x)+', '+
                       str(y)+', '+
                       str(circle_radius)+') # color='+circle_color+' width=2 \n')

        if (number is None) and (value is None):
            reg_syntax2 = ('text( '+
                           str(x)+', '+
                           str(y+y_desc)+', '+
                           '{'+'detected source'+'})'+
                           ' # color='+circle_color+' width=2 \n')
        elif number is None:
            reg_syntax2 = ('text( '+
                           str(x)+', '+
                           str(y+y_desc)+', '+
                           '{'+str(round(value, 1))+'})'+
                           ' # color='+circle_color+' width=2 \n')
        elif value is None:
            reg_syntax2 = ('text( '+
                           str(x)+', '+
                           str(y+y_desc)+', '+
                           '{'+str(number)+'})'+
                           ' # color='+circle_color+' width=2 \n')
        else:
            reg_syntax2 = ('text( '+
                           str(x)+', '+
                           str(y+y_desc)+', '+
                           '{'+str(number)+','+str(round(value, 1))+'})'+
                           ' # color='+circle_color+' width=2 \n')

        return reg_syntax1, reg_syntax2



class NormedPSF(object):
    """
    Attributes
    ----------
    fits : astropy.io.fits.hdu.image.PrimaryHDU
    object_df : pandas.core.frame.DataFrame (SourceExtractor.dataframe)
        DataFrame of object data from SExtractor
    psf_size : int, default 14
        Length of ArtificialReal PSF map : psf_size*2 + 1

    sky_value : numpy.float64 (float)
    values : numpy.ndarray

    Examples
    --------
    >>> normed_psf = NormedPSF(astropy_fits_hdu[0], pandas_dataframe)

    """
    def __init__(self, fits, object_df, psf_size=14):
        """
        Parameters
        ----------
        fits : astropy.io.fits.hdu.image.PrimaryHDU
        object_df : pandas.core.frame.DataFrame (SourceExtractor.dataframe)
            DataFrame of object data from SExtractor
        psf_size : int, default 14
            Length of ArtificialReal PSF map : psf_size*2 + 1

        """
        self.fits      = fits
        self.object_df = object_df
        self.psf_size  = psf_size

        self._calc_normed_psf()

    def _calc_normed_psf(self):
        psf_maps = []
        sky_values = []
        for index, pd_series in self.object_df.iterrows():
            x0 = int( math.floor(pd_series['X_IMAGE'] + 0.5) - 1)
            y0 = int( math.floor(pd_series['Y_IMAGE'] + 0.5) - 1)
            l  = self.psf_size
            cutout_map = self.fits.data[y0-l-1:y0+l,x0-l-1:x0+l]

            try: # test!
                maximum    = np.max(cutout_map)
            except:
                print x0, y0, cutout_map.shape, self.fits.data.shape
                sys.exit('Occured Exception.')

            aperture_outside_list = [self.fits.data[y][x]
                                     for x in range(x0-l-1,x0+l) for y in range(y0-l-1,y0+l)
                                     if (x-x0)**2 + (y-y0)**2 > l**2 ]
            sky_value = np.median(aperture_outside_list)

            object_psf = (cutout_map-sky_value)/(maximum-sky_value)
            
            psf_maps.append(object_psf)
            sky_values.append(sky_value)

        normed_psf_map = np.median(np.array(psf_maps), axis=0) / np.median(np.array(psf_maps), axis=0).max()
        normed_psf_map = np.where(normed_psf_map<0, 0, normed_psf_map)

        self.sky_value = np.median(sky_values)
        self.values = normed_psf_map



class ArtificialReal(object):
    """
    Attributes
    ----------
    input_fits : astropy.io.fits.hdu.image.PrimaryHDU
    psf : numpy.ndarray
        Require shape : (n,n)
    fits_edge : int, default 30

    Examples
    --------
    >>> input_fits = astropy.io.fits.open(fits_path)[0]
    >>> artificial = ArtificialReal(input_fits, psf, psf_size=14)
    >>> embed_dataframe = artificial.make_random(embed_num = 200,
                                                 flux_peak_minpow = -2.0,
                                                 flux_peak_maxpow = 3.0)
    >>> artificial.embed_fits(embed_dataframe, output_fits_name='embed_random.fits')



    >>> embed_dataframe = artificial.make_around_galaxy(galaxy_dataframe,
                                                        flux_peak_minpow = -2.0,
                                                        flux_peak_maxpow = 3.0)
    >>> artificial.embed_fits(embed_dataframe, output_fits_name='embed_around_galaxy.fits')
    """
    def __init__(self, input_fits, psf):
        """
        Parameters
        ----------
        input_fits : astropy.io.fits.hdu.image.PrimaryHDU
        psf : numpy.ndarray
            Require shape : (n,n)
        """
        self.input_fits = input_fits
        self.psf        = psf
        self.fits_edge  = 30 # Avoid edges of image

    def make_random(self,
                    embed_num=100,
                    flux_peak_minpow=-2.0,
                    flux_peak_maxpow=3.0):
        embed_list = [[random.uniform(0+self.fits_edge,self.input_fits.header['NAXIS1']-self.fits_edge),
                       random.uniform(0+self.fits_edge,self.input_fits.header['NAXIS2']-self.fits_edge),
                       10**random.uniform(flux_peak_minpow, flux_peak_maxpow)] for i in range(embed_num)]
        embed_dataframe = pd.DataFrame(embed_list, columns=['X_IMAGE','Y_IMAGE','FLUX_MAX'])
        return embed_dataframe

    def make_around_galaxy(self, galaxy_dataframe,
                           flux_peak_minpow=-2.0,
                           flux_peak_maxpow=3.0,
                           max_sigma=3):
        def _make_around_galaxy_core(galaxy_series):
            x,y = ArtificialReal._unit_circle_random_uniform()
            theta = galaxy_series['THETA_IMAGE']
            x *= max_sigma * galaxy_series['A_IMAGE']
            y *= max_sigma * galaxy_series['B_IMAGE']
            x1 = x*np.cos(theta) - y*np.sin(theta)
            y1 = x*np.sin(theta) + y*np.cos(theta)
            embed_x = galaxy_series['X_IMAGE']+x1
            embed_y = galaxy_series['Y_IMAGE']+y1
            flux_max = 10**random.uniform(flux_peak_minpow, flux_peak_maxpow)
            return [ embed_x, embed_y, flux_max ]

        embed_list = [ _make_around_galaxy_core(galaxy_series) for index, galaxy_series in galaxy_dataframe.iterrows() if (0+self.fits_edge+max_sigma*galaxy_series['A_IMAGE'] < galaxy_series['X_IMAGE'] < self.input_fits.header['NAXIS1']-self.fits_edge-max_sigma*galaxy_series['A_IMAGE']) and (0+self.fits_edge+max_sigma*galaxy_series['A_IMAGE'] < galaxy_series['Y_IMAGE'] < self.input_fits.header['NAXIS2']-self.fits_edge-max_sigma*galaxy_series['A_IMAGE']) ]
        embed_dataframe = pd.DataFrame(embed_list, columns=['X_IMAGE','Y_IMAGE','FLUX_MAX'])
        return embed_dataframe

    def embed_fits(self, embed_dataframe, output_fits_name='embed.fits'):
        output_fits_data = self.input_fits.data
        for index, embed_series in embed_dataframe.iterrows():
            x_min = int(embed_series['X_IMAGE'] - self.psf.shape[0]/2) - 1
            x_max = int(embed_series['X_IMAGE'] + self.psf.shape[0]/2) - 1 + 1
            y_min = int(embed_series['Y_IMAGE'] - self.psf.shape[1]/2) - 1
            y_max = int(embed_series['Y_IMAGE'] + self.psf.shape[1]/2) - 1 + 1
            output_fits_data[y_min:y_max,x_min:x_max] += self.psf * embed_series['FLUX_MAX']

        hdu = astropy.io.fits.PrimaryHDU(output_fits_data, self.input_fits.header)
        hdu_list = astropy.io.fits.HDUList([hdu])
        hdu_list.writeto(output_fits_name, overwrite=True)

    @staticmethod
    def _unit_circle_random_uniform():
        radius_max = 1.
        radius = np.sqrt(2*random.uniform(0., 0.5*radius_max*radius_max))
        theta = random.uniform(0., 2.*np.pi)
        x = radius*np.cos(theta)
        y = radius*np.sin(theta)
        return x, y



class FilterSource(object):

    def __init__(self, ref_fits_path, ref_remap_fits_path, sub_fits_path, psf_size=14):
        self.ref_hdu = astropy.io.fits.open(ref_fits_path)
        self.ref_remap_hdu = astropy.io.fits.open(ref_remap_fits_path)
        self.sub_hdu = astropy.io.fits.open(sub_fits_path)
        self.psf_size = psf_size
        self.ref_wcs = astropy.wcs.WCS(self.ref_hdu[0].header)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.ref_hdu.close()
        self.ref_remap_hdu.close()
        self.sub_hdu.close()

    def check_non_ref(self, pd_series):
        flag = False
        x0_image = pd_series['X_IMAGE']
        y0_image = pd_series['Y_IMAGE']
        x1 = int(x0_image) - self.psf_size
        if x1 < 0 : x1 = 0
        x2 = int(x0_image) + self.psf_size + 1
        if x2 > self.sub_hdu[0].header['NAXIS1'] : x2 = self.sub_hdu[0].header['NAXIS1']
        y1 = int(y0_image) - self.psf_size
        if y1 < 0 : y1 = 0
        y2 = int(y0_image) + self.psf_size + 1
        if y2 > self.sub_hdu[0].header['NAXIS2'] : y2 = self.sub_hdu[0].header['NAXIS2']
        cutout_image = self.ref_remap_hdu[0].data[y1:y2,x1:x2]
        if np.any(cutout_image == 0): flag = True
        return flag

    def check_saturation(self, pd_series):
        flag = False
        thresh_value_1 = -5
        thresh_value_2 = 1e-4
        l = 3
        ra = pd_series['ALPHA_J2000']
        dec = pd_series['DELTA_J2000']
        pixel_index = self.ref_wcs.wcs_world2pix([[ra, dec]],0)
        x0_image, y0_image = pixel_index[0]
        x1 = int(x0_image) - l
        if x1 < 0 : x1 = 0
        x2 = int(x0_image) + l + 1
        if x2 > self.ref_hdu[0].header['NAXIS1'] : x2 = self.ref_hdu[0].header['NAXIS1']
        y1 = int(y0_image) - l
        if y1 < 0 : y1 = 0
        y2 = int(y0_image) + l + 1
        if y2 > self.ref_hdu[0].header['NAXIS2'] : y2 = self.ref_hdu[0].header['NAXIS2']
        cutout_image = self.ref_hdu[0].data[y1:y2,x1:x2]

        if np.any(cutout_image == 0): flag = True # 1st criteria
        for rad in range(cutout_image.shape[0]/2):
            if FilterSource.check_saturation_edge(cutout_image,rad,thresh=thresh_value_1) and (np.any(abs(cutout_image) < thresh_value_2) or np.any(cutout_image < 0)):
                flag = True # 2nd criteria
        return flag

    @staticmethod
    def check_saturation_edge(image,radius,thresh=-1):
        flag = False
        x1 = image.shape[0]/2 - radius
        y1 = image.shape[1]/2 - radius
        x2 = x1 - 1
        y2 = y1 - 1
        north_inside  = image[x1:-x1,y1:y1+1]
        south_inside  = image[x1:-x1,-(y1+1):-y1]
        west_inside   = image[x1:x1+1,y1:-y1]
        east_inside   = image[-(x1+1):-x1,y1:-y1]
        north_outside = image[x1:-x1,y2:y2+1]
        south_outside = image[x1:-x1,-(y2+1):-y2]
        west_outside  = image[x2:x2+1,y1:-y1]
        east_outside  = image[-(x2+1):-x2,y1:-y1]
        if np.any(north_inside - north_outside < thresh): flag = True
        if np.any(south_inside - south_outside < thresh): flag = True
        if np.any(west_inside - west_outside < thresh): flag = True
        if np.any(east_inside - east_outside < thresh): flag = True
        return flag

    def check_mask(self, pd_series):
        flag = False
        mask_value = 1.e-20
        x0_image = pd_series['X_IMAGE']
        y0_image = pd_series['Y_IMAGE']
        x1 = int(x0_image) - self.psf_size
        if x1 < 0 : x1 = 0
        x2 = int(x0_image) + self.psf_size + 1
        if x2 > self.sub_hdu[0].header['NAXIS1'] : x2 = self.sub_hdu[0].header['NAXIS1']
        y1 = int(y0_image) - self.psf_size
        if y1 < 0 : y1 = 0
        y2 = int(y0_image) + self.psf_size + 1
        if y2 > self.sub_hdu[0].header['NAXIS2'] : y2 = self.sub_hdu[0].header['NAXIS2']
        cutout_image = self.sub_hdu[0].data[y1:y2,x1:x2]
        if np.any(np.abs(cutout_image) <= mask_value): flag = True
        return flag

    def check_edge(self, pd_series):
        flag = False
        edge = 30
        x0_image = pd_series['X_IMAGE']
        y0_image = pd_series['Y_IMAGE']
        if(x0_image < edge): flag = True
        if(y0_image < edge): flag = True
        if(self.sub_hdu[0].header['NAXIS1'] - x0_image < edge): flag = True
        if(self.sub_hdu[0].header['NAXIS2'] - y0_image < edge): flag = True
        return flag



class Cutout(object):

    def __init__(self, ref_fits_path, new_fits_path, sub_fits_path):
        self.ref_hdu = astropy.io.fits.open(ref_fits_path)
        self.new_hdu = astropy.io.fits.open(new_fits_path)
        self.sub_hdu = astropy.io.fits.open(sub_fits_path)

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, trace):
        self.ref_hdu.close()
        self.new_hdu.close()
        self.sub_hdu.close()

    @classmethod
    def cutout_image(cls, x0_image, y0_image, image, psf_size=14):
        l = psf_size
        x0_image = int( math.floor(x0_image + 0.5) - 1)
        y0_image = int( math.floor(y0_image + 0.5) - 1)
        psf_nd   = image[(y0_image-l):(y0_image+l+1),(x0_image-l):(x0_image+l+1)]
        return psf_nd



def remove_files(path_name):
    for path in glob.glob(path_name):
        if os.path.isfile(path):
            os.remove(path)



def move_files(path_name, dir_path):
    command_mv = 'mv ' + path_name + ' ' + dir_path
    os.system(command_mv)



def templete_image(connection, RAcen, DECcen, match_rad_temp=1.0, directory_temp_path='/lustre/tomoesn/reference'):
    # match_rad_temp: matching radius (1.0 deg)    
    with connection.cursor() as cur:
        query = 'SELECT "fitsName",q3c_dist(ra, dec, {}, {}) AS dist, abs(ra- {}) AS absra, abs(dec- {}) AS absdec FROM template WHERE q3c_radial_query(ra, dec, {},  {}, {}) ORDER BY absdec ASC, absra ASC LIMIT 1'.format(RAcen, DECcen, RAcen, DECcen, RAcen, DECcen, match_rad_temp)
        try:
            cur.execute(query)
        except:
            time.sleep(10)
            cur.execute(query)
        row = cur.fetchone()
    temp_fits_path = os.path.join(directory_temp_path, row[0]) # ps1 image
    dist = row[1]
    return temp_fits_path, dist



def imsub(connection, input_fits_path, temp_remap_fits_name='temp_remap.fits', subtracted_fits_name='sub.fits'):

    flag = 0
    
    # image subtraction
    # for input
    iu = 45000.0 # upper limit 
    il = -500.0 # lower limit 
    ig = 1.0 # gain 
    ir = 1.0 # readout noise
    # for tempate
    tu = iu # upper limit
    tl = il # lower limit
    tg = ig # gain
    tr = ir # readout noise

    # get header info
    input_fits_hdu = astropy.io.fits.open(input_fits_path)
    input_fits = input_fits_hdu[0]
    NAXIS1 = input_fits.header['NAXIS1']
    NAXIS2 = input_fits.header['NAXIS2']
    CTYPE1 = input_fits.header['CTYPE1']
    CTYPE2 = input_fits.header['CTYPE2']
    CRVAL1 = input_fits.header['CRVAL1']
    CRVAL2 = input_fits.header['CRVAL2']
    CRPIX1 = input_fits.header['CRPIX1']
    CRPIX2 = input_fits.header['CRPIX2']
    CD1_1  = input_fits.header['CD1_1']
    CD1_2  = input_fits.header['CD1_2']
    CD2_1  = input_fits.header['CD2_1']
    CD2_2  = input_fits.header['CD2_2']

    if(CTYPE2.find("DEC") != -1): # x <=> RA, y <=> DEC
        #print "  x~RA, y~DEC"
        RAcen  = CRVAL1 + CD1_1*(NAXIS1/2-CRPIX1) + CD1_2*(NAXIS2/2-CRPIX2)
        DECcen = CRVAL2 + CD2_1*(NAXIS1/2-CRPIX1) + CD2_2*(NAXIS2/2-CRPIX2)
    else: # y <=> RA, x <=> DEC
        #print "  x~DEC, y-RA"
        DECcen = CRVAL1 + CD1_1*(NAXIS1/2-CRPIX1) + CD1_2*(NAXIS2/2-CRPIX2)
        RAcen  = CRVAL2 + CD2_1*(NAXIS1/2-CRPIX1) + CD2_2*(NAXIS2/2-CRPIX2)

    # Template image
    temp_fits_path, _ = templete_image(connection, RAcen, DECcen)
    if(os.path.exists(temp_fits_path) == 0):
        #print 'Error! template image does not exist!'
        flag = 1
        return flag, None

    # Remap template image (wcsremap)
    command_wcsremap ="wcsremap -template " + input_fits_path + " -source " + temp_fits_path + " -outIm " + temp_remap_fits_name + " >/dev/null 2>&1"
    os.system(command_wcsremap)

    # image subtraction (hotpants)
    command_hotpants = ("hotpants -inim " + input_fits_path + " -tmplim " + temp_remap_fits_name + " -outim " + subtracted_fits_name +
                        " -iu " + str(iu) + " -il " + str(il) + " -ig " + str(ig) + " -ir " + str(ir) +
                        " -tu " + str(tu) + " -tl " + str(tl) + " -tg " + str(tg) + " -tr " + str(tr) + " >/dev/null 2>&1")
    os.system(command_hotpants)

    return flag, temp_fits_path



if __name__=='__main__':
    SetWorkEnv_str = '  Checked module.'
    print SetWorkEnv_str
